/*********************************************************/
/* A program that displays using a fairly simplistic     */
/*     'stochastic' transparency technique.  In quotes   */
/*     since a positional hash is used for randomization */
/* This demonstrates the use of IGLU render-to-texture,  */
/*     multisample framebuffers, shaders, trackball,     */
/*     vertex arrays, matrices, and use of multiple      */
/*     shaders, changing shader variables during display */
/*     and perhaps a few other things.                   */
/* The original non-IGLU demo by Ethan Kerzner.          */
/*********************************************************/

#include <GL/glew.h>
#include <stdio.h>

#include "iglu.h"
using namespace iglu;

// How many samples are in our multisample buffer?
#define NUM_SAMPLES_PER_PIXEL  32

// What types of drawing techniques will we use?
enum MODES
{
	DRAW_TO_TEXTURE = 0,
	DRAW_TO_SCREEN,
	NUM_DRAW_MODES
};

// A selector to allow us to swap between drawing to a 32-sample
//    framebuffer object or the (1-sample) screen
int                     drawMode  = DRAW_TO_TEXTURE;

// Key matriices variable for this program
IGLUMatrix4x4           modelview;

// Shaders used in this sample program
IGLUShaderProgram::Ptr  resolve            = 0;
IGLUShaderProgram::Ptr  alphaWeightedCoin  = 0;

// An intermediate framebuffer object we'll use to store a multisample buffer
IGLUFramebuffer::Ptr    msFbo     = 0;

// A simple trackball to rotate our geometry
IGLUTrackball::Ptr		ball = 0;

// A vertex array to store our geometry
IGLUVertexArray::Ptr    vertArray = 0;

// A frame timer we'll use to display a counter on screen
IGLUFrameRate::Ptr      frameTime = 0;

// Data for a simple square
float vertData[20] = { 
	0, 1,  -1, -1, 0, 
	1, 1,  1, -1, 0, 
	0, 0,  -1, 1, 0,
	1, 0,  1, 1, 0, 
};


void DrawSquares( void )
{
	alphaWeightedCoin->Enable();
	
		// Draw the first square.  White, opaque.
		alphaWeightedCoin["model"] = modelview * ball->GetMatrix();;
		alphaWeightedCoin["color"] = vec4( 1.0f, 1.0f, 1.0f, 1.0f );
		vertArray->DrawArrays( GL_TRIANGLE_STRIP, 0, 4 );
		
		// Draw the second square.  Red, 50% opaque.
		alphaWeightedCoin["model"] = modelview * ball->GetMatrix() * IGLUMatrix4x4::Translate(1.0f, 0.0f, 1.0f);
		alphaWeightedCoin["color"] = vec4( 1.0f, 0.0f, 0.0f, 0.5f );
		vertArray->DrawArrays( GL_TRIANGLE_STRIP, 0, 4 );

		// Draw the third square.  Blue, 25% opaque.
		alphaWeightedCoin["model"] = modelview * ball->GetMatrix() * IGLUMatrix4x4::Translate(-0.5f, 1.0f, 0.1f);
		alphaWeightedCoin["color"] = vec4( 0.0f, 0.0f, 1.0f, 0.25 );
		vertArray->DrawArrays( GL_TRIANGLE_STRIP, 0, 4 );
	
	alphaWeightedCoin->Disable();
}

void display ( void )
{
	// Start timing this frame
	frameTime->StartFrame();

	// Clear the screen to start the frame fresh
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	// Pick one of our two drawing modes to use
	if (drawMode == DRAW_TO_TEXTURE)
	{
		// Draw the squares into our 32-sample buffer
		msFbo->Bind();
		msFbo->Clear();
			DrawSquares( );
		msFbo->Unbind();

		// Do a resolve phase to figure out the color based on all 32 samples
		resolve->Enable();
			vertArray->DrawArrays( GL_TRIANGLE_STRIP, 0, 4 );
		resolve->Disable();
	}
	else if (drawMode == DRAW_TO_SCREEN)
		// Draw the squares into out 1-sample screen buffer
		DrawSquares( );	

	// Draw a framerate counter on screen
	char buf[32];
	sprintf( buf, "%.1f fps", frameTime->EndFrame(), 0.5 );
	IGLUDraw::DrawText( IGLU_FONT_FIXED, 0, 0, buf );

	// Swap the GLUT buffers so we're actually show what we drew.
	glutSwapBuffers();
}

void mouse( int button, int state, int x, int y )
{
	// Handle trackball clicks and releases
	if ( state == GLUT_DOWN )
		ball->SetOnClick( x, y );
	else if (state == GLUT_UP )
		ball->Release();
}

void motion( int x, int y )
{
	// Update the trackball matrix when we detect motion.
	ball->UpdateOnMotion( x, y ); 
	glutPostRedisplay();
}


void keys(unsigned char key, int x, int y)
{
	switch(key)
	{
	case 27:
	case'q':
	case'Q':
		exit(0);
		break;
	case'a':
		drawMode = (drawMode+1) % NUM_DRAW_MODES; 
		printf("Mode: Rendering %s\n",
		       drawMode == DRAW_TO_TEXTURE ? 
	               "squares to texture then to screen" :
		           "squares to screen");
		break;
	}
}

void idle ( void )
{
	glutPostRedisplay();
}

int main(int argc, char** argv)
{
	glutInit(&argc, argv);
	glutInitDisplayMode( GLUT_RGBA | GLUT_DOUBLE | GLUT_MULTISAMPLE);
	glutInitWindowSize( 1024, 768 );
	glutInitWindowPosition( 20, 20 );
	glutCreateWindow( "More Complex IGLU Example:  Really Simple 'Stochastic' Transparency" );
	glewInit();
	glutDisplayFunc( display );
	glutKeyboardFunc( keys );
	glutIdleFunc( idle );
	glutMouseFunc( mouse );
	glutMotionFunc( motion );

	// Create our frame timer
	frameTime = new IGLUFrameRate( 20 );

	// Create a virtual trackball
	printf("(+) Creating a virtual trackball...\n");
	ball = new IGLUTrackball( 1024, 768 );

	printf("(+) Initializing modelview matrix...\n");
	modelview  = IGLUMatrix4x4::LookAt(vec3(1.0f, 0.0f, 5.0f), vec3(0.0f, 0.0f, 0.0f), vec3(0.0f, 1.0f, 0.0f));
	
	printf("(+) Creating multisample framebuffer...\n");
	msFbo = IGLUFramebuffer::CreateMultisample( GL_RGBA32F, 1024, 768, NUM_SAMPLES_PER_PIXEL, true );

	printf("(+) Loading vertex and fragment shaders...\n" );
	
	// First, load the shader that takes geometry and draws it into a multisample buffer
	alphaWeightedCoin  = new IGLUShaderProgram( "shaders/textureDisplay.vert.glsl", "shaders/solidColor.frag.glsl");
	alphaWeightedCoin->SetProgramEnables( IGLU_GLSL_DEPTH_TEST | IGLU_GLSL_MULTISAMPLE | IGLU_GLSL_SAMPLE_SHADING );
	alphaWeightedCoin["proj"] = IGLUMatrix4x4::Perspective(60, 1024.0 / 768.0, 0.1, 10.0);

	// Now load the shader that takes the multisample buffer and figures out a color to display on screen
	resolve                 = new IGLUShaderProgram( "shaders/textureDisplay.vert.glsl", "shaders/resolve.frag.glsl");
	resolve->SetProgramEnables( IGLU_GLSL_DEPTH_TEST | IGLU_GLSL_MULTISAMPLE ); 
	resolve["theRenderTex"] = msFbo[0];
	resolve["proj"]         = IGLUMatrix4x4::Identity();   
	resolve["model"]        = IGLUMatrix4x4::Identity();

	printf("(+) Initializing geometry...\n");
	vertArray = new IGLUVertexArray();
	vertArray->SetVertexArray( sizeof( vertData ), vertData, IGLU_STATIC|IGLU_DRAW );

	// Tell OpenGL where data from this vertex array gets piped into the shaders.
	//    -> Since the layout(location = x) is the same in both shaders this does not
	//       need to be done everytime the current shader is changed.
	vertArray->EnableAttribute( resolve["vertex"], 3, GL_FLOAT, 5 * sizeof(float), BUFFER_OFFSET(2*sizeof(float)) );
	vertArray->EnableAttribute( resolve["texcoord"], 2, GL_FLOAT, 5 * sizeof(float), BUFFER_OFFSET(0) );

	// Start the GLUT interaction loop
	glutMainLoop();
	return 0;
}
